#!/usr/bin/python

import os
import sys
import gettext
PROGNAME="redhat-config-bind"
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)

try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode    

if not "/usr/lib/python2.2/site-packages/rhpl" in sys.path:
    sys.path.append("/usr/lib/python2.2/site-packages/rhpl")

from Conf import *
TRUE=1
FALSE=not TRUE

hname_re = re.compile('^[a-zA-Z0-9.\-]+$')

ip_re = re.compile('^([0-2]?[0-9]?[0-9])\\.([0-2]?[0-9]?[0-9])\\.([0-2]?[0-9]?[0-9])\\.([0-2]?[0-9]?[0-9])(?:\\.in-addr\\.arpa\\.?)?$')

revip_re = re.compile('^([0-2]?[0-9]?[0-9])(?:\\.([0-2]?[0-9]?[0-9]))?(?:\\.([0-2]?[0-9]?[0-9])\\.)?(?:\\.([0-2]?[0-9]?[0-9]))?(?:\\.in-addr\\.arpa)$')


def testName(value):
    if not hname_re.match(value):
        raise TestError, _("Zone name %s is not a valid domain name.") % value
		
class TestError(Exception):
	def __init__(self, args=None):
		#raise Error("Here with args" + args)
		self.args = args

	#def __str__(self):
		#return Exception.__str__(self)

def testAppliesTo(value):
    pass

def testFile(value):
    pass

def testHost(value):
    if not (hname_re.match(value) or value == '@'):
        raise TestError, _("%s is not a valid hostname.") % value
    
def testIp(value):
    if not checkIpNum(value):
        raise TestError, _("%s is not a valid IP address.") % value

def checkIpNum(value):
	try:
		m = ip_re.match(value)
		if not m:
			return FALSE
		nums = m.groups()
		if not nums or ( nums and ( len(nums) != 4 )) :
			return FALSE
		
		for i in xrange(0, 4):
			if nums[i] == None:
				return FALSE
			num = int(nums[i])
			if (num < 0) or (num > 255):
				return FALSE
	except TypeError:
		return FALSE
	
	return TRUE


def checkRevIpNum(value):
	try:
		m = revip_re.match(value)
		if not m:
			return FALSE
		nums = m.groups()
		
		if not nums: return FALSE
		
		#print len(nums), " matching groups!", nums
		for i in xrange(0, len(nums)):
			if nums[i] == None:
				continue
			num = int(nums[i])
			if (num < 0) or (num > 255):
				return FALSE
	except TypeError:
		return FALSE

	return TRUE

def checkTTL(ttl):
	if ttl < 0:
		return FALSE
	return TRUE
	# add more TTL checking code here

# Zone
#  This class presents a data-oriented class for making changes
#  to the /var/named/ Zone files
class SOA:
    def __init__(self,pns="@",server="localhost",contact="root"):
        self.pns=pns
        self.server=server
        self.contact=contact
        self.SOA=[1,28800,14400,3600000,86400]
        
    def load(self,SOAList):
        self.SOA=SOAList
        
    def getPNS(self):
        return self.pns
    def setPNS(self,pns):
        self.pns=pns

    def getTTL(self):
        return self.SOA[4]
    def setTTL(self,ttl):
        self.SOA[4]=ttl

    def getRefresh(self):
        return self.SOA[1]
    def setRefresh(self,refresh):
        self.SOA[1]=refresh

    def getRetry(self):
        return self.SOA[2]
    def setRetry(self,retry):
        self.SOA[2]=retry

    def getExpire(self):
        return self.SOA[3]
    def setExpire(self,expire):
        self.SOA[3]=expire

    def getSerial(self):
        return self.SOA[0]
    def setSerial(self,i):
        self.SOA[0]=i
    def incSerial(self):
        self.SOA[0]=self.SOA[0]+1

    def setContact(self,contact):
        self.contact=contact
    def getContact(self):
        return self.contact
    def setServer(self,server):
        self.server=server
    def getServer(self):
        return self.server

    def testPNS(self, value):
        if not value:
            raise TestError, _("Primary Nameserver (SOA) not defined.")
		
        if ((len(value) == 1) and value == "@"):
            return

        if not ((len(value) > 1) and value[-1] == '.'):
            raise TestError, _("Primary Nameserver (SOA) '%s' has no . at the end. You must use a full hostname.") %  value
        if not hname_re.match(value):
            raise TestError, _("Primary Nameserver (SOA) '%s' is not a valid hostname or IP address.") % value
		
    def testSerial(self, value):
        if value <= 0:
            raise TestError, _("Serial number must be >= 1.")
        
    def testRefresh(self, value):
        if not checkTTL(value):
            raise TestError, _("Refresh value must be >= 0.")

    def testRetry(self, value):
        if not checkTTL(value):
            raise TestError, _("Retry value must be >= 0.")

    def testExpire(self, value):
        if not checkTTL(value):
            raise TestError, _("Expire value must be >= 0.")

    def testTTL(self, value):
        if not checkTTL(value):
            raise TestError, _("TTL value must be >= 0.")

    def out(self):
        ret="%s\tIN\tSOA\t%s\t%s\t(\n" % (self.getPNS(),self.getServer(),self.getContact())
        ret="%s\t\t\t\t%d ; serial\n" % (ret,self.getSerial())
        ret="%s\t\t\t\t%d ; refresh\n" % (ret,self.getRefresh())
        ret="%s\t\t\t\t%d ; retry\n" % (ret,self.getRetry())
        ret="%s\t\t\t\t%d ; expire\n" % (ret,self.getExpire())
        ret="%s\t\t\t\t%d ; ttl\n" % (ret,self.getTTL())
        ret="%s\t\t\t\t)\n\n\n" % (ret)
        return ret

    
class A:
    def __init__(self,host=None,ip=None,ttl="",parent=None):
        self.host=host
        self.ip=ip
        self.parent=parent
        self.ttl=ttl
    def getHost(self):
        return self.host
    def setHost(self,host):
        self.host=host
    def getName(self):
        return self.getHost()
    def getIp(self):
        return self.ip
    def setIp(self,ip):
        self.ip=ip
    def unlink(self):
        self.parent.unlink(self)
    def getTtl(self):
        return self.ttl
    def setTtl(self,ttl):
        self.ttl=ttl

    def testHost(self, value):
        testHost(value)

    def testIp(self, value):
        testIp(value)

    def get_str(self):
        return _("host '%s'") % self.getHost()
		
    def out(self):
        return "%s\t%s\tIN\tA\t%s" % (self.host,self.ttl,self.ip)

class PTR:
    def __init__(self,host,ip,parent=None):
        self.host=host
        self.ip=ip
        self.parent=parent
    def getHost(self):
        return self.host
    def setHost(self,host):
        self.host=host
    def getName(self):
        return self.getHost()
    def getIp(self):
        return self.ip
    def setIp(self,ip):
        self.ip=ip
    def unlink(self):
        self.parent.unlink(self)

    def out(self):
        return "%s\t\tIN\tPTR\t%s" % (self.ip,self.host)

class CNAME:
    def __init__(self,host,alias,ttl,parent):
        self.host=host
        self.alias=alias
        self.parent=parent
        self.ttl=ttl
    def getHost(self):
        return self.host
    def setHost(self,host):
        self.host=host
    def getName(self):
        return self.getHost()
    def getAlias(self):
        return self.alias
    def setAlias(self,alias):
        self.alias=alias
    def getTtl(self):
        return self.ttl
    def setTtl(self,ttl):
        self.ttl=ttl
    def unlink(self):
        self.parent.unlink(self)
    def out(self):
        return "%s\t%s\tIN CNAME\t%s" % (self.alias,self.ttl,self.host)

    def testAlias(self, value):
        testHost(value)

class SRV:
    def __init__(self,host,service,port,parent):
        self.host=host
        self.service=service
        self.parent=parent
        self.port=port
    def getHost(self):
        return self.host
    def setHost(self,host):
        self.host=host
    def getName(self):
        return self.getService()
    def getService(self):
        return self.service
    def setService(self,service):
        self.service=service

    def getPort(self):
        return self.port
    def setPort(self,port):
        self.port=port
        
    def getWeight(self):
        return self.weight
    def setWeight(self,weight):
        self.weight=weight
        
    def getTTL(self):
        return self.TTL
    def setTTL(self,TTL):
        self.TTL=TTL
        
    def getPriority(self):
        return self.Priority
    def setPriority(self,priority):
        self.priority=priority
        
    def unlink(self):
        self.parent.unlink(self)
    def out(self):
        return "%s\t\t%s %s %s IN SRV\t%s %s %s" % (self.service,self.getTTL(),getpriority(),self.getWeight(),self.getPort(),self.getHost())

class MX:
    def __init__(self,name,priority,server,ttl,parent=None):
        self.name=name # val[0]
        self.priority=priority #atoi(val[3])
        self.server=server #val[4]
        self.parent=parent
        self.ttl=ttl
    def getServer(self):
        return self.server
    def setServer(self,server):
        self.server=server
    def setName(self,name):
        self.name=name
    def getName(self):
        return self.name
    def getTtl(self):
        return self.ttl
    def setTtl(self,ttl):
        self.ttl=ttl
    def getPriority(self):
        return self.priority
    def setPriority(self,priority):
        self.priority=priority
    def unlink(self):
        self.parent.unlink(self)
    def out(self):
        return "%s\t%s\tIN\tMX\t%d\t%s" % (self.name,self.ttl,self.priority,self.server)

class NS:
    def __init__(self,applies_to,host,parent):
        self.fullhost=host
        self.applies_to=applies_to
        self.parent=parent
    def getHost(self):
        return self.fullhost.split(".")[0]
    def setHost(self,host):
        self.host=host
    def getName(self):
        return self.getHost()
    def getFullHost(self):
        return self.fullhost
    def setAppliesTo(self,applies_to):
        self.applies_to=applies_to
    def getAppliesTo(self):
        return self.applies_to
    def unlink(self):
        self.parent.unlink(self)
    def out(self):
        return "%s\tIN\tNS\t%s" % (self.getAppliesTo(),self.getHost())
        
class Zone(Conf):
    def __init__(self,name,filename):
	Conf.__init__(self, filename, commenttype=';' )
        self.SOA=SOA()
        self.name=name
        self.read()
        self.rewind()
        self.zones=[]
        self.origins=[]
        current=[]
        prev_val="@"
        self.NS=[]
        self.A=[]
        self.PTR=[]
        self.MX=[]
        self.CNAME=[]
        self.SRV=[]
        while self.findnextcodeline():
            line=self.getline().split(";")
            val=line[0].split()
            self.nextline()
            if val[0]=="$ORIGIN":
                self.origins.append(val)
                continue
            if val[0]=="$TTL":
                self.TTL=val[1]
                continue
            if "SOA" in val:
                i=val.index("SOA")
                if val[0]!="IN":
                    pns=val[0]
                else:
                    pns=""
                self.SOA=SOA(pns,val[i+1],val[i+2])
                done=FALSE
                val=val[i+3:]
                SOAList=[]
                while not done:
                    if len(val)==0 or val[0]==";":
                        val=self.getline().split()
                        self.findnextcodeline()
                        self.nextline()
                    if val[0]==")":
                        done=TRUE
                        break
                    if val[0]=="(":
                        val=val[1:]
                        continue
                    SOAList.append(self.translate(val[0]))
                    if val[0][-1]==")":
                        done=TRUE
                        break
                    val=val[1:]

                self.SOA.load(SOAList)
                continue
            if "NS" in val:
                if val[0]=="IN":
                    applies_to="@"
                else:
                    applies_to=val[0]
            
                self.addNS(val[-1],applies_to)
                continue
            if "A" in val:
                if val[0] != "IN" and not val[0].isdigit():
                    prev_val=val[0]
                self.addA(prev_val,val)
                continue
            if "MX" in val:
                if val[0] != "IN" and not val[0].isdigit():
                    prev_val=val[0]
                self.addMX(val,prev_val)
                continue

            if "SRV" in val:
                self.addSRV(val[0],val[-2],val[-1])
                continue
            
            if "CNAME" in val:
                if val[0] != "CNAME":
                    prev_val=val[0]
                self.addCNAME(val)
                continue

            if "PTR" in val:
                self.addPTR(val[-1],val[0])
                continue
            
            print "unknown",val

    def translate(self,val):
        if val[-1]=="D":
            return atoi(val[:-1])*60*60*24
        if val[-1]=="H":
            return atoi(val[:-1])*60*60
        if val[-1]=="W":
            return atoi(val[:-1])*60*60*24*7
        if val[-1]=="M":
            return atoi(val[:-1])*60
        if val[-1]==")":
            return atoi(val[:-1])
        return atoi(val)
    def unlink(self,rec):
        if isinstance(rec,A):
            self.A.remove(rec)
        if isinstance(rec,NS):
            self.NS.remove(rec)
        if isinstance(rec,MX):
            self.MX.remove(rec)
        if isinstance(rec,PTR):
            self.PTR.remove(rec)
        if isinstance(rec,CNAME):
            self.CNAME.remove(rec)
            
    def getName(self):
        return self.name
    def isReverse(self):
        x=self.filename.split(".in-addr.arpa")
        return self.filename!=x[0]

    def add(self,name,ip):
        if self.isReverse():
            for i in self.PTR:
                if i.getIp()==ip:
                    raise str(_("%s already exists") % ip)
            else:
                self.addPTR(ip,name)
        else:
            for a in self.A:
                if a.getName()==name:
                    raise str(_("%s already exists") % name)
            else:
                val=[name,ip]
                self.addA(name,val)
        
    def modify(self,name,ip):
        if self.isReverse():
            for i in self.PTR:
                if i.getIp()==ip:
                    self.addPTR(ip,name)
                    return
            raise str(_("%s does not exist") % ip)
        else:
            for n in self.A:
                if n.getName()==name:
                    val=[name,ip]
                    self.addA(name,val)
                    return
            raise str(_("%s does not exist") % name)
        
    def setServer(self,server):
        self.SOA.setServer(server)
    def getServer(self):
        return self.SOA.getServer()

    def setContact(self,contact):
        self.SOA.setContact(contact)
    def getContact(self):
        return self.SOA.getContact()

    def getSOA(self):
        return self.SOA

    def getTTL(self):
        return self.SOA.getTTL()
    def setTTL(self,ttl):
        self.setTTL(ttl)

    def getRefresh(self):
        return self.SOA.getRefresh()
    def setRefresh(self,refresh):
        self.SOA.setRefresh(refresh)

    def getRetry(self):
        return self.SOA.getRetry()
    def setRetry(self,retry):
        self.SOA.setRetry(retry)

    def getExpire(self):
        return self.SOA.getExpire()
    def setExpire(self,expire):
        self.SOA.setExpire(expire)

    def getSerial(self):
        return self.SOA.getSerial()
    def setSerial(self,i):
        self.SOA.setSerial(i)
    def incSerial(self):
        self.SOA.incSerial()

    def getNSList(self):
        return self.NS

    def addNS(self,name,applies_to):
        self.NS.append(NS(applies_to,name,self))

    def getNS(self,val):
        return self.NS

    def __outNS(self):
        ret=""
        for i in self.NS:
            ret="%s%s\n" % (ret,i.out())
        return ret


    def getAList(self):
        return self.A

    def addA(self,name,val):
        ttl=""
        if len(val) > 2:
            ttl=val[1]
        self.A.append(A(name,val[-1],ttl,self))
                      
    def getA(self,host):
        for i in self.A:
            if i.getHost()==host:
                return i
        return None
    def __outA(self):
        ret=""
        for i in self.A:
            ret="%s%s\n" % (ret,i.out())
        return ret

    def addPTR(self,name,ip):
        self.PTR.append(PTR(name,ip,self))
        
    def getPTRList(self):
        return self.PTR

    def __outPTR(self):
        ret=""
        for i in self.PTR:
            ret="%s%s\n" % (ret,i.out())
        return ret

    def getMXList(self):
        return self.MX

    def addMX(self,val,prev_val):
        ttl=""
        if val[0].isdigit():
            ttl=val[0]
        if val[1].isdigit():
            ttl=val[0]
        self.MX.append(MX(prev_val,atoi(val[-2]),val[-1],ttl,self))

    def getMX(self,val):
        for i in self.MX:
            if i.getName()==name:
                return self.i
        return None
    def __outMX(self):
        ret=""
        for i in self.MX:
            ret="%s%s\n" % (ret,i.out())
        return ret

    def getCNAMEList(self):
        return self.CNAME

    def addCNAME(self,val):
        if len(val) == 5:
            ttl=val[1]
        else:
            ttl=""
        self.CNAME.append(CNAME(val[-1],val[0],ttl,self))
        
    def getCNAME(self,val):
        for c in self.CNAME:
            if c.getHost()==val:
                return c
        return None

    def getSRVList(self):
        return self.SRV

    def addSRV(self,service,port,host):
        self.SRV.append(SRV(service,port,host,self))
        
    def getSRV(self,val):
        for c in self.SRV:
            if c.getService()==val:
                return c
        return None

    def save(self,uid,gid):
        self.incSerial()
        tmpFile="%s.%d" % (self.filename,os.getpid())
        fd = open(tmpFile,"w")
        fd.write(self.out())
        fd.close()
        os.rename(tmpFile,self.filename)
        os.chown(self.filename,uid,gid)

    def __outCNAMES(self):
        ret=""
        for i in self.CNAME:
            ret="%s%s\n" % (ret,i.out())
        return ret

    def findnextcommentline(self):
        # optional whitespace followed by non-comment character
        # defines a codeline.  blank lines, lines with only whitespace,
        # and comment lines do not count.
        return self.findnextline('^[' + self.commenttype + ']+')
    
    def outComments(self):
        self.read()
        self.rewind()
        ret=""
        while self.findnextcommentline():
            val=self.getline()
            ret="%s%s\n" % (ret,val)
            self.nextline()
        return ret

        
    def out(self):
        ret=self.outComments()
        for origin in self.origins:
            ret="%s%s\n" % (ret,join(origin,' '))
        ret="%s$TTL %s\n" % (ret,self.getTTL())
        ret="%s%s\n" % (ret,self.SOA.out())
        ret="%s%s\n" % (ret,self.__outNS())
        ret="%s%s\n" % (ret,self.__outA())
        ret="%s%s\n" % (ret,self.__outCNAMES())
        ret="%s%s\n" % (ret,self.__outMX())
        ret="%s%s\n" % (ret,self.__outPTR())
        return ret

def Usage():
    print _("Zone - Python named zone commandline tool\n\nUsage: zone [-a|-m|-d|-l] [-h] [-i <IP> ] [-n <Name> ] name")

if __name__ == "__main__":
    import signal
    import getopt

    if os.getuid() != 0:
        print _("Please restart %s with root permissions!") % (sys.argv[0])
        sys.exit(10)
        
    if len(sys.argv) == 1:
        Usage()
        sys.exit(1)
        
    cmdline = sys.argv[1:]
    sys.argv = sys.argv[:1]

    signal.signal (signal.SIGINT, signal.SIG_DFL)
    class BadUsage: pass
    
    zoneName = cmdline[-1]
    ip=None
    name=None
    type="A"
    add=FALSE
    modify=FALSE
    delete=FALSE
    progname = os.path.basename(sys.argv[0])

    try:
        opts, args = getopt.getopt(cmdline, "amdlhi:n:t:", ["add","modify","delete", "help","ip=","name=","list","type"])
        for opt, val in opts:
            
            if opt == '-a' or opt == '--add':
                add=TRUE

            if opt == '-m' or opt == '--modify':
                modify=TRUE

            if opt == '-d' or opt == '--delete':
                delete=TRUE

            if opt == '-i' or opt == '--ip':
                ip=val

            if opt == '-p' or opt == '--protocol':
                protocol=val

            if opt == '-n' or opt == '--name':
                name=val

            if opt == '-t' or opt == '--type':
                type=val

            if opt == '-l' or opt == '--list':
                z=Zone(zoneName,zoneName)
                print z.out()
                sys.exit(0)

            if opt == '-h' or opt == '--help':
                Usage()
                sys.exit(0)

    except (getopt.error, BadUsage):
        Usage()
        sys.exit(1)

    if (add and delete) or (add and modify) or (modify and delete):
        Usage()
        sys.exit(1)

    try:
        z=Zone(zoneName, zoneName)
        if delete:
            z.delete(name)
            sys.exit(0)

        if add or modify:
            if add:
                z.add(name,ip)

            if modify:
                z.modify(name,ip)

            z.save()
            sys.exit(0)

        found=FALSE
        print z.getOpt(name)
                
    except ValueError,e:
        print _("Error: %s") % e
        sys.exit(1)
        
